Package: SBOAtools
Type: Package
Title: Secretary Bird Optimization for Continuous Optimization and
        Neural Network Training
Version: 0.1.1
Authors@R: c(
    person("Burak", "Dilber", email = "burakdilber91@gmail.com", role = c("aut", "cre", "cph")),
    person("A. Fırat", "Özdemir", email = "firat.ozdemir5761@gmail.com", role = c("aut", "cph"))
    )
Description: Provides an implementation of Secretary Bird Optimization for
    general-purpose continuous optimization, benchmark optimization, and
    training single-hidden-layer feed-forward neural network models. The
    implemented optimizer is based on the Secretary Bird Optimization Algorithm
    proposed by Fu et al. (2024) <doi:10.1007/s10462-024-10729-y>. The neural
    network training functionality is based on Dilber and Özdemir (2026)
    <doi:10.1007/s00521-026-11874-x>.
License: MIT + file LICENSE
URL: https://github.com/burakdilber/SBOAtools
BugReports: https://github.com/burakdilber/SBOAtools/issues
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: stats, graphics
NeedsCompilation: no
Packaged: 2026-05-02 15:20:24 UTC; hp
Author: Burak Dilber [aut, cre, cph],
  A. Fırat Özdemir [aut, cph]
Maintainer: Burak Dilber <burakdilber91@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-03 06:30:17 UTC
Built: R 4.5.2; ; 2026-05-03 08:28:14 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,sboa)
S3method(plot,sboa_mlp)
S3method(predict,sboa_mlp)
S3method(print,sboa)
S3method(print,sboa_mlp)
export(get_benchmark)
export(list_benchmarks)
export(sboa)
export(sboa_mlp)
importFrom(stats,rnorm)
importFrom(stats,runif)

# SBOAtools

**SBOAtools** is an R package developed for the **Secretary Bird Optimization Algorithm (SBOA)**. The package supports both **general-purpose continuous optimization** and **single-hidden-layer multilayer perceptron (MLP) training**.

It is intended for researchers working in **metaheuristic optimization**, **computational intelligence**, and **neural network training**. The package allows users to apply SBOA either as a standalone optimizer or as a training algorithm for feed-forward neural networks.

---

## Features

- General-purpose continuous optimization with `sboa()`
- Single-hidden-layer MLP training with `sboa_mlp()`
- Prediction support via `predict()`
- Convergence visualization via `plot()`
- Model summaries via `print()`

---

## Installation

During development, the package can be installed from the local source using:

```r
devtools::install()
```

Then load the package with:

```r
library(SBOAtools)
```

---

You can also install the development version from GitHub:

```r
install.packages("remotes")
remotes::install_github("burakdilber/SBOAtools")
```

---

## Main Functions

### `sboa()`

Performs general-purpose continuous optimization using the Secretary Bird Optimization Algorithm.

### `sboa_mlp()`

Trains a single-hidden-layer multilayer perceptron using the Secretary Bird Optimization Algorithm.

---

## Example 1: General Optimization

```r
library(SBOAtools)

sphere <- function(x) sum(x^2)

res <- sboa(
  fn = sphere,
  lower = rep(-10, 5),
  upper = rep(10, 5),
  n_agents = 10,
  max_iter = 20,
  seed = 123
)

print(res)
plot(res)

res$value
res$par
```

---

## Example 2: MLP Training with SBOA

```r
library(SBOAtools)

set.seed(123)

X_train <- matrix(runif(40), nrow = 10, ncol = 4)
y_train <- matrix(runif(10), nrow = 10, ncol = 1)

fit_mlp <- sboa_mlp(
  X_train = X_train,
  y_train = y_train,
  hidden_dim = 3,
  n_agents = 10,
  max_iter = 20,
  lower = -1,
  upper = 1,
  seed = 123
)

print(fit_mlp)
plot(fit_mlp)

pred <- predict(fit_mlp, X_train)
pred
```

---

## Returned Objects

### Output of `sboa()`

The `sboa()` function returns an object of class `"sboa"` containing:

- `par`: best solution found
- `value`: best objective function value
- `convergence`: convergence curve over iterations
- `population`: final population matrix
- `fitness`: final fitness values of the population
- `call`: matched function call

### Output of `sboa_mlp()`

The `sboa_mlp()` function returns an object of class `"sboa_mlp"` containing:

- `par`: optimized neural network parameters
- `value`: best objective function value
- `convergence`: convergence curve over iterations
- `input_dim`: number of input variables
- `hidden_dim`: number of hidden neurons
- `output_dim`: number of output variables
- `x_min`: minimum values used for input normalization
- `x_max`: maximum values used for input normalization
- `y_min`: minimum values used for output normalization
- `y_max`: maximum values used for output normalization
- `fitted`: fitted values on the original scale
- `metrics`: training performance metrics
- `call`: matched function call

---

## Current Scope

The current version of the package supports:

- continuous optimization problems
- single-hidden-layer MLP models
- sigmoid activation in hidden and output layers
- regression-oriented neural network training with mean squared error (MSE)

---

## Future Extensions

Possible future improvements include:

- additional benchmark functions
- engineering optimization examples
- train/test evaluation helpers
- classification support
- alternative activation functions
- multiple hidden layers
- visualization utilities

---

## Authors

- **Burak Dilber**
- **A. Fırat Özdemir**

---

## References

- Fu, W., Wang, K., Liu, J., et al. (2024). *Secretary Bird Optimization Algorithm*. **Artificial Intelligence Review**. https://doi.org/10.1007/s10462-024-10729-y

- Dilber, B., & Ozdemir, A. F. (2026). *A novel approach to training feed-forward multi-layer perceptrons with recently proposed secretary bird optimization algorithm*. **Neural Computing and Applications**. https://doi.org/10.1007/s00521-026-11874-x

## License

MIT License

# SBOAtools

**SBOAtools** is an R package for the **Secretary Bird Optimization Algorithm (SBOA)**. The package supports both **general-purpose continuous optimization** and **single-hidden-layer multilayer perceptron (MLP) training**.

It is designed for researchers working in **metaheuristic optimization**, **computational intelligence**, and **neural network training**. The package allows users to apply SBOA either as a standalone optimizer or as a training algorithm for feed-forward neural networks.

---

## Features

- General-purpose continuous optimization with `sboa()`
- Single-hidden-layer MLP training with `sboa_mlp()`
- Built-in benchmark functions (`F1`-`F23`)
- Benchmark inspection with `list_benchmarks()`
- Benchmark retrieval with `get_benchmark()`
- Prediction support via `predict()`
- Convergence visualization via `plot()`
- Model summaries via `print()`

---

## Installation

During development, the package can be installed from the local source using:

```r
devtools::install()
```

Then load the package with:

```r
library(SBOAtools)
```

You can also install the development version from GitHub:

```r
install.packages("remotes")
remotes::install_github("burakdilber/SBOAtools")
```

---

## Main Functions

### `sboa()`

Performs general-purpose continuous optimization using the Secretary Bird Optimization Algorithm.

### `sboa_mlp()`

Trains a single-hidden-layer multilayer perceptron using the Secretary Bird Optimization Algorithm.

### `list_benchmarks()`

Displays the built-in benchmark functions available in the package.

### `get_benchmark()`

Returns a benchmark definition and its metadata.

---

## Built-in Benchmark Functions

`SBOAtools` includes 23 built-in benchmark functions (`F1`-`F23`) for continuous optimization studies.

You can inspect the available benchmark functions with:

```r
list_benchmarks()
```

You can retrieve a benchmark definition and its metadata with:

```r
b <- get_benchmark("F9")
b$label
b$category
b$fn(rep(1, 5))
```

The built-in benchmark set includes unimodal, multimodal, and fixed-dimension functions.

---

## Example 1: General Optimization with a User-Defined Function

```r
library(SBOAtools)

sphere <- function(x) sum(x^2)

res <- sboa(
  fn = sphere,
  lower = rep(-10, 5),
  upper = rep(10, 5),
  n_agents = 10,
  max_iter = 20,
  seed = 123,
  verbose = FALSE
)

print(res)
plot(res)

res$value
res$par
```

---

## Example 2: General Optimization with a Built-in Benchmark

```r
library(SBOAtools)

list_benchmarks()

res2 <- sboa(
  fn = "F1",
  lower = rep(-100, 30),
  upper = rep(100, 30),
  n_agents = 30,
  max_iter = 500,
  seed = 123,
  verbose = FALSE
)

print(res2)
plot(res2)
```

You can also inspect a specific benchmark before optimization:

```r
b <- get_benchmark("F14")
b$label
b$fixed_dim
```

---

## Example 3: MLP Training with SBOA

```r
library(SBOAtools)

set.seed(123)

X_train <- matrix(runif(40), nrow = 10, ncol = 4)
y_train <- matrix(runif(10), nrow = 10, ncol = 1)

fit_mlp <- sboa_mlp(
  X_train = X_train,
  y_train = y_train,
  hidden_dim = 3,
  n_agents = 10,
  max_iter = 20,
  lower = -1,
  upper = 1,
  seed = 123,
  verbose = FALSE
)

print(fit_mlp)
plot(fit_mlp)

pred <- predict(fit_mlp, X_train)
pred
```

---

## Returned Objects

### Output of `sboa()`

The `sboa()` function returns an object of class `"sboa"` containing:

- `par`: best solution found
- `value`: best objective function value
- `convergence`: convergence curve over iterations
- `population`: final population matrix
- `fitness`: final fitness values of the population
- `call`: matched function call

### Output of `sboa_mlp()`

The `sboa_mlp()` function returns an object of class `"sboa_mlp"` containing:

- `par`: optimized neural network parameters
- `value`: best objective function value
- `convergence`: convergence curve over iterations
- `input_dim`: number of input variables
- `hidden_dim`: number of hidden neurons
- `output_dim`: number of output variables
- `x_min`: minimum values used for input normalization
- `x_max`: maximum values used for input normalization
- `y_min`: minimum values used for output normalization
- `y_max`: maximum values used for output normalization
- `fitted`: fitted values on the original scale
- `metrics`: training performance metrics
- `call`: matched function call

---

## Current Scope

The current version of the package supports:

- continuous optimization problems
- built-in benchmark functions for optimization experiments
- single-hidden-layer MLP models
- sigmoid activation in hidden and output layers
- regression-oriented neural network training with mean squared error (MSE)

---

## Future Extensions

Possible future improvements include:

- engineering optimization examples
- train/test evaluation helpers
- classification support
- alternative activation functions
- multiple hidden layers
- additional visualization utilities
- automatic benchmark bound selection within `sboa()`

---

## Authors

- **Burak Dilber**
- **A. Fırat Özdemir**

---

## References

- Fu, W., Wang, K., Liu, J., et al. (2024). *Secretary Bird Optimization Algorithm*. **Artificial Intelligence Review**. https://doi.org/10.1007/s10462-024-10729-y

- Dilber, B., & Ozdemir, A. F. (2026). *A novel approach to training feed-forward multi-layer perceptrons with recently proposed secretary bird optimization algorithm*. **Neural Computing and Applications**. https://doi.org/10.1007/s00521-026-11874-x

---

## License

MIT License

#' Train a single-hidden-layer MLP using SBOA
#'
#' Trains a single-hidden-layer multilayer perceptron with the
#' Secretary Bird Optimization Algorithm.
#'
#' @param X_train Training input data.
#' @param y_train Training output data.
#' @param hidden_dim Number of hidden neurons.
#' @param n_agents Number of search agents.
#' @param max_iter Maximum number of iterations.
#' @param lower Lower bound for parameter search.
#' @param upper Upper bound for parameter search.
#' @param seed Optional random seed.
#' @param verbose Logical; if \code{TRUE}, progress is printed.
#'
#' @return An object of class \code{"sboa_mlp"}.
#' @references
#' Dilber, B., and Ozdemir, A. F. (2026). A novel approach to training
#' feed-forward multi-layer perceptrons with recently proposed secretary bird
#' optimization algorithm. Neural Computing and Applications.
#' DOI: 10.1007/s00521-026-11874-x
#' @examples
#' set.seed(123)
#'
#' X_train <- matrix(runif(40), nrow = 10, ncol = 4)
#' y_train <- matrix(runif(10), nrow = 10, ncol = 1)
#'
#' fit <- sboa_mlp(
#'   X_train = X_train,
#'   y_train = y_train,
#'   hidden_dim = 3,
#'   n_agents = 10,
#'   max_iter = 20,
#'   lower = -1,
#'   upper = 1,
#'   seed = 123,
#'   verbose = FALSE
#' )
#'
#' print(fit)
#' pred <- predict(fit, X_train)
#' head(pred)
#' @export
sboa_mlp <- function(X_train,
                     y_train,
                     hidden_dim = 10,
                     n_agents = 30,
                     max_iter = 500,
                     lower = -1,
                     upper = 1,
                     seed = NULL,
                     verbose = TRUE) {

  X_train <- as.matrix(X_train)
  y_train <- as.matrix(y_train)

  if (!is.null(seed)) {
    set.seed(seed)
  }

  x_norm <- normalize_minmax(X_train)
  y_norm <- normalize_minmax(y_train)

  Xs <- x_norm$x_scaled
  Ys <- y_norm$x_scaled

  input_dim <- ncol(Xs)
  output_dim <- ncol(Ys)

  n_params <- input_dim * hidden_dim +
    hidden_dim +
    hidden_dim * output_dim +
    output_dim

  opt <- sboa(
    fn = mlp_mse_fitness,
    lower = rep(lower, n_params),
    upper = rep(upper, n_params),
    n_agents = n_agents,
    max_iter = max_iter,
    X = Xs,
    Y = Ys,
    input_dim = input_dim,
    hidden_dim = hidden_dim,
    output_dim = output_dim,
    verbose = verbose,
    seed = seed
  )

  fitted_scaled <- forward_mlp(
    params = opt$par,
    X = Xs,
    input_dim = input_dim,
    hidden_dim = hidden_dim,
    output_dim = output_dim
  )$Y_hat

  fitted <- denormalize_minmax(fitted_scaled, y_norm$min, y_norm$max)

  metrics <- list(
    RMSE = rmse_vec(y_train, fitted),
    MAE  = mae_vec(y_train, fitted),
    MAPE = mape_vec(y_train, fitted),
    RSQ  = rsq_vec(y_train, fitted)
  )

  out <- list(
    par = opt$par,
    value = opt$value,
    convergence = opt$convergence,
    input_dim = input_dim,
    hidden_dim = hidden_dim,
    output_dim = output_dim,
    x_min = x_norm$min,
    x_max = x_norm$max,
    y_min = y_norm$min,
    y_max = y_norm$max,
    fitted = fitted,
    metrics = metrics,
    call = match.call()
  )

  class(out) <- "sboa_mlp"
  out
}

#' Predict method for SBOA-MLP objects
#'
#' Generates predictions from a trained SBOA-MLP model.
#'
#' @param object An object of class \code{"sboa_mlp"}.
#' @param newdata New predictor data.
#' @param ... Additional arguments, ignored.
#'
#' @return A matrix of predicted values.
#' @export
predict.sboa_mlp <- function(object, newdata, ...) {
  newdata <- as.matrix(newdata)
  newdata_scaled <- apply_minmax(newdata, object$x_min, object$x_max)

  pred_scaled <- forward_mlp(
    params = object$par,
    X = newdata_scaled,
    input_dim = object$input_dim,
    hidden_dim = object$hidden_dim,
    output_dim = object$output_dim
  )$Y_hat

  denormalize_minmax(pred_scaled, object$y_min, object$y_max)
}

#' Print method for SBOA-MLP objects
#'
#' Prints a summary of a trained SBOA-MLP model.
#'
#' @param x An object of class \code{"sboa_mlp"}.
#' @param ... Additional arguments, ignored.
#'
#' @return The input object, invisibly.
#' @export
print.sboa_mlp <- function(x, ...) {
  cat("SBOA-MLP model\n")
  cat("Input dimension :", x$input_dim, "\n")
  cat("Hidden dimension:", x$hidden_dim, "\n")
  cat("Output dimension:", x$output_dim, "\n")
  cat("Best objective  :", signif(x$value, 6), "\n")
  cat("RMSE            :", signif(x$metrics$RMSE, 6), "\n")
  cat("MAE             :", signif(x$metrics$MAE, 6), "\n")
  cat("MAPE            :", signif(x$metrics$MAPE, 6), "\n")
  cat("R-squared       :", signif(x$metrics$RSQ, 6), "\n")
  invisible(x)
}

#' Plot method for SBOA-MLP objects
#'
#' Plots the convergence curve of a trained SBOA-MLP model.
#'
#' @param x An object of class \code{"sboa_mlp"}.
#' @param ... Additional graphical arguments passed to \code{plot()}.
#'
#' @return No return value. Called for its side effect.
#' @export
plot.sboa_mlp <- function(x, ...) {
  plot(
    x$convergence,
    type = "l",
    xlab = "Iteration",
    ylab = "Best fitness",
    main = "SBOA-MLP Convergence",
    ...
  )
}

#' Root mean squared error
#'
#' Computes the root mean squared error between actual and predicted values.
#'
#' @param actual Numeric vector or matrix of observed values.
#' @param predicted Numeric vector or matrix of predicted values.
#'
#' @return A numeric scalar.
rmse_vec <- function(actual, predicted) {
  actual <- as.numeric(actual)
  predicted <- as.numeric(predicted)
  sqrt(mean((actual - predicted)^2))
}

#' Mean absolute error
#'
#' Computes the mean absolute error between actual and predicted values.
#'
#' @param actual Numeric vector or matrix of observed values.
#' @param predicted Numeric vector or matrix of predicted values.
#'
#' @return A numeric scalar.
mae_vec <- function(actual, predicted) {
  actual <- as.numeric(actual)
  predicted <- as.numeric(predicted)
  mean(abs(actual - predicted))
}

#' Mean absolute percentage error
#'
#' Computes the mean absolute percentage error between actual and predicted values.
#'
#' @param actual Numeric vector or matrix of observed values.
#' @param predicted Numeric vector or matrix of predicted values.
#' @param eps Small constant to avoid division by zero.
#'
#' @return A numeric scalar.
mape_vec <- function(actual, predicted, eps = 1e-8) {
  actual <- as.numeric(actual)
  predicted <- as.numeric(predicted)
  mean(abs((actual - predicted) / pmax(abs(actual), eps))) * 100
}

#' R-squared
#'
#' Computes the coefficient of determination.
#'
#' @param actual Numeric vector or matrix of observed values.
#' @param predicted Numeric vector or matrix of predicted values.
#'
#' @return A numeric scalar.
rsq_vec <- function(actual, predicted) {
  actual <- as.numeric(actual)
  predicted <- as.numeric(predicted)

  ss_res <- sum((actual - predicted)^2)
  ss_tot <- sum((actual - mean(actual))^2)

  1 - ss_res / ss_tot
}

#' Sigmoid activation function
#'
#' Computes the sigmoid transformation.
#'
#' @param x A numeric vector, matrix, or scalar.
#'
#' @return Transformed values in the interval (0, 1).
sigmoid <- function(x) {
  1 / (1 + exp(-x))
}

#' Clip values to lower and upper bounds
#'
#' Restricts each element of a vector to the given bounds.
#'
#' @param x Numeric vector.
#' @param lower Lower bound vector.
#' @param upper Upper bound vector.
#'
#' @return A bounded numeric vector.
clip_bounds <- function(x, lower, upper) {
  pmin(pmax(x, lower), upper)
}

#' Lévy flight step generator
#'
#' Generates Lévy-distributed random steps.
#'
#' @param d Dimension of the step vector.
#' @param beta Stability parameter.
#'
#' @return A numeric vector of length \code{d}.
#' @importFrom stats rnorm
levy_flight <- function(d, beta = 1.5) {
  sigma <- (gamma(1 + beta) * sin(pi * beta / 2) /
              (gamma((1 + beta) / 2) * beta * 2^((beta - 1) / 2)))^(1 / beta)
  u <- rnorm(d) * sigma
  v <- rnorm(d)
  u / abs(v)^(1 / beta)
}

#' Min-max normalization
#'
#' Scales each column of a matrix to the interval from 0 to 1.
#'
#' @param x Numeric vector or matrix.
#'
#' @return A list containing scaled data and min-max values.
normalize_minmax <- function(x) {
  x <- as.matrix(x)
  mins <- apply(x, 2, min)
  maxs <- apply(x, 2, max)
  ranges <- maxs - mins
  ranges[ranges == 0] <- 1

  x_scaled <- sweep(x, 2, mins, "-")
  x_scaled <- sweep(x_scaled, 2, ranges, "/")

  list(
    x_scaled = x_scaled,
    min = mins,
    max = maxs
  )
}

#' Apply min-max normalization using existing bounds
#'
#' @param x Numeric vector or matrix.
#' @param mins Column minima.
#' @param maxs Column maxima.
#'
#' @return Scaled matrix.
apply_minmax <- function(x, mins, maxs) {
  x <- as.matrix(x)
  ranges <- maxs - mins
  ranges[ranges == 0] <- 1

  x_scaled <- sweep(x, 2, mins, "-")
  sweep(x_scaled, 2, ranges, "/")
}

#' Reverse min-max normalization
#'
#' @param x_scaled Scaled numeric vector or matrix.
#' @param mins Column minima.
#' @param maxs Column maxima.
#'
#' @return Values on the original scale.
denormalize_minmax <- function(x_scaled, mins, maxs) {
  x_scaled <- as.matrix(x_scaled)
  ranges <- maxs - mins
  sweep(sweep(x_scaled, 2, ranges, "*"), 2, mins, "+")
}

#' Print method for SBOA objects
#'
#' Prints a short summary of an SBOA optimization result.
#'
#' @param x An object of class \code{"sboa"}.
#' @param ... Additional arguments, ignored.
#'
#' @return The input object, invisibly.
#' @export
print.sboa <- function(x, ...) {
  cat("Secretary Bird Optimization Algorithm (SBOA)\n")
  cat("Best objective value:", signif(x$value, 6), "\n")
  cat("Dimension:", length(x$par), "\n")
  cat("Number of iterations:", length(x$convergence), "\n")
  invisible(x)
}

#' Plot method for SBOA objects
#'
#' Plots the convergence curve of an SBOA optimization result.
#'
#' @param x An object of class \code{"sboa"}.
#' @param ... Additional graphical arguments passed to \code{plot()}.
#'
#' @return No return value. Called for its side effect.
#' @export
plot.sboa <- function(x, ...) {
  plot(
    x$convergence,
    type = "l",
    xlab = "Iteration",
    ylab = "Best objective value",
    main = "SBOA Convergence Curve",
    ...
  )
}

# Internal helper: penalty function used in F12 and F13
.u_penalty <- function(x, a, k, m) {
  ifelse(
    x > a, k * (x - a)^m,
    ifelse(x < -a, k * (-x - a)^m, 0)
  )
}

# ---------------------------------
# Built-in benchmark functions F1-F23
# ---------------------------------

.f1 <- function(x) {
  x <- as.numeric(x)
  sum(x^2)
}

.f2 <- function(x) {
  x <- as.numeric(x)
  sum(abs(x)) + prod(abs(x))
}

.f3 <- function(x) {
  x <- as.numeric(x)
  sum(cumsum(x)^2)
}

.f4 <- function(x) {
  x <- as.numeric(x)
  max(abs(x))
}

.f5 <- function(x) {
  x <- as.numeric(x)
  sum(100 * (x[-1] - x[-length(x)]^2)^2 + (x[-length(x)] - 1)^2)
}

.f6 <- function(x) {
  x <- as.numeric(x)
  sum((floor(x + 0.5))^2)
}

.f7 <- function(x) {
  x <- as.numeric(x)
  i <- seq_along(x)
  sum(i * x^4) + runif(1)
}

.f8 <- function(x) {
  x <- as.numeric(x)
  sum(-x * sin(sqrt(abs(x))))
}

.f9 <- function(x) {
  x <- as.numeric(x)
  n <- length(x)
  10 * n + sum(x^2 - 10 * cos(2 * pi * x))
}

.f10 <- function(x) {
  x <- as.numeric(x)
  n <- length(x)
  -20 * exp(-0.2 * sqrt(sum(x^2) / n)) -
    exp(sum(cos(2 * pi * x)) / n) + 20 + exp(1)
}

.f11 <- function(x) {
  x <- as.numeric(x)
  i <- seq_along(x)
  sum(x^2) / 4000 - prod(cos(x / sqrt(i))) + 1
}

.f12 <- function(x) {
  x <- as.numeric(x)
  n <- length(x)
  y <- 1 + (x + 1) / 4

  term1 <- 10 * sin(pi * y[1])^2
  term2 <- sum((y[1:(n - 1)] - 1)^2 * (1 + 10 * sin(pi * y[2:n])^2))
  term3 <- (y[n] - 1)^2

  (pi / n) * (term1 + term2 + term3) + sum(.u_penalty(x, 10, 100, 4))
}

.f13 <- function(x) {
  x <- as.numeric(x)
  n <- length(x)

  term1 <- sin(3 * pi * x[1])^2
  term2 <- sum((x[1:(n - 1)] - 1)^2 * (1 + sin(3 * pi * x[2:n])^2))
  term3 <- (x[n] - 1)^2 * (1 + sin(2 * pi * x[n])^2)

  0.1 * (term1 + term2 + term3) + sum(.u_penalty(x, 5, 100, 4))
}

.f14 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 2) {
    stop("F14 requires a 2-dimensional input.")
  }

  a1 <- c(-32, -16, 0, 16, 32)
  aS <- rbind(rep(a1, 5), rep(a1, each = 5))

  s <- numeric(25)
  for (j in 1:25) {
    s[j] <- sum((x - aS[, j])^6)
  }

  1 / (1 / 500 + sum(1 / ((1:25) + s)))
}

.f15 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 4) {
    stop("F15 requires a 4-dimensional input.")
  }

  aK <- c(
    0.1957, 0.1947, 0.1735, 0.16,
    0.0844, 0.0627, 0.0456, 0.0342,
    0.0323, 0.0235, 0.0246
  )
  bK <- 1 / c(
    0.25, 0.5, 1, 2,
    4, 6, 8, 10,
    12, 14, 16
  )

  sum((aK - (x[1] * (bK^2 + x[2] * bK)) /
         (bK^2 + x[3] * bK + x[4]))^2)
}

.f16 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 2) {
    stop("F16 requires a 2-dimensional input.")
  }

  4 * x[1]^2 - 2.1 * x[1]^4 + x[1]^6 / 3 +
    x[1] * x[2] - 4 * x[2]^2 + 4 * x[2]^4
}

.f17 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 2) {
    stop("F17 requires a 2-dimensional input.")
  }

  a <- 1
  b <- 5.1 / (4 * pi^2)
  c <- 5 / pi
  r <- 6
  s <- 10
  t <- 1 / (8 * pi)

  a * (x[2] - b * x[1]^2 + c * x[1] - r)^2 + s * (1 - t) * cos(x[1]) + s
}

.f18 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 2) {
    stop("F18 requires a 2-dimensional input.")
  }

  term1 <- 1 + (x[1] + x[2] + 1)^2 *
    (19 - 14 * x[1] + 3 * x[1]^2 - 14 * x[2] + 6 * x[1] * x[2] + 3 * x[2]^2)

  term2 <- 30 + (2 * x[1] - 3 * x[2])^2 *
    (18 - 32 * x[1] + 12 * x[1]^2 + 48 * x[2] - 36 * x[1] * x[2] + 27 * x[2]^2)

  term1 * term2
}

.f19 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 3) {
    stop("F19 requires a 3-dimensional input.")
  }

  alpha <- c(1.0, 1.2, 3.0, 3.2)

  A <- rbind(
    c(3.0, 10, 30),
    c(0.1, 10, 35),
    c(3.0, 10, 30),
    c(0.1, 10, 35)
  )

  P <- 1e-4 * rbind(
    c(3689, 1170, 2673),
    c(4699, 4387, 7470),
    c(1091, 8732, 5547),
    c(381, 5743, 8828)
  )

  outer_sum <- numeric(4)
  for (i in 1:4) {
    outer_sum[i] <- sum(A[i, ] * (x - P[i, ])^2)
  }

  -sum(alpha * exp(-outer_sum))
}

.f20 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 6) {
    stop("F20 requires a 6-dimensional input.")
  }

  alpha <- c(1.0, 1.2, 3.0, 3.2)

  A <- rbind(
    c(10, 3, 17, 3.5, 1.7, 8),
    c(0.05, 10, 17, 0.1, 8, 14),
    c(3, 3.5, 1.7, 10, 17, 8),
    c(17, 8, 0.05, 10, 0.1, 14)
  )

  P <- 1e-4 * rbind(
    c(1312, 1696, 5569, 124, 8283, 5886),
    c(2329, 4135, 8307, 3736, 1004, 9991),
    c(2348, 1451, 3522, 2883, 3047, 6650),
    c(4047, 8828, 8732, 5743, 1091, 381)
  )

  outer_sum <- numeric(4)
  for (i in 1:4) {
    outer_sum[i] <- sum(A[i, ] * (x - P[i, ])^2)
  }

  -sum(alpha * exp(-outer_sum))
}

.f21 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 4) {
    stop("F21 requires a 4-dimensional input.")
  }

  aSH <- rbind(
    c(4, 4, 4, 4),
    c(1, 1, 1, 1),
    c(8, 8, 8, 8),
    c(6, 6, 6, 6),
    c(3, 7, 3, 7),
    c(2, 9, 2, 9),
    c(5, 5, 3, 3),
    c(8, 1, 8, 1),
    c(6, 2, 6, 2),
    c(7, 3.6, 7, 3.6)
  )

  cSH <- c(0.1, 0.2, 0.2, 0.4, 0.4, 0.6, 0.3, 0.7, 0.5, 0.5)

  m <- 5
  -sum(1 / (rowSums((aSH[1:m, , drop = FALSE] - matrix(x, nrow = m, ncol = 4, byrow = TRUE))^2) + cSH[1:m]))
}

.f22 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 4) {
    stop("F22 requires a 4-dimensional input.")
  }

  aSH <- rbind(
    c(4, 4, 4, 4),
    c(1, 1, 1, 1),
    c(8, 8, 8, 8),
    c(6, 6, 6, 6),
    c(3, 7, 3, 7),
    c(2, 9, 2, 9),
    c(5, 5, 3, 3),
    c(8, 1, 8, 1),
    c(6, 2, 6, 2),
    c(7, 3.6, 7, 3.6)
  )

  cSH <- c(0.1, 0.2, 0.2, 0.4, 0.4, 0.6, 0.3, 0.7, 0.5, 0.5)

  m <- 7
  -sum(1 / (rowSums((aSH[1:m, , drop = FALSE] - matrix(x, nrow = m, ncol = 4, byrow = TRUE))^2) + cSH[1:m]))
}

.f23 <- function(x) {
  x <- as.numeric(x)
  if (length(x) != 4) {
    stop("F23 requires a 4-dimensional input.")
  }

  aSH <- rbind(
    c(4, 4, 4, 4),
    c(1, 1, 1, 1),
    c(8, 8, 8, 8),
    c(6, 6, 6, 6),
    c(3, 7, 3, 7),
    c(2, 9, 2, 9),
    c(5, 5, 3, 3),
    c(8, 1, 8, 1),
    c(6, 2, 6, 2),
    c(7, 3.6, 7, 3.6)
  )

  cSH <- c(0.1, 0.2, 0.2, 0.4, 0.4, 0.6, 0.3, 0.7, 0.5, 0.5)

  m <- 10
  -sum(1 / (rowSums((aSH[1:m, , drop = FALSE] - matrix(x, nrow = m, ncol = 4, byrow = TRUE))^2) + cSH[1:m]))
}

# ---------------------------------
# Benchmark registry
# ---------------------------------

.benchmark_registry <- function() {
  list(
    F1 = list(
      name = "F1",
      label = "Sphere",
      category = "unimodal",
      fn = .f1,
      lower = -100,
      upper = 100,
      fixed_dim = NA_integer_
    ),
    F2 = list(
      name = "F2",
      label = "Schwefel 2.22",
      category = "unimodal",
      fn = .f2,
      lower = -10,
      upper = 10,
      fixed_dim = NA_integer_
    ),
    F3 = list(
      name = "F3",
      label = "Schwefel 1.2",
      category = "unimodal",
      fn = .f3,
      lower = -100,
      upper = 100,
      fixed_dim = NA_integer_
    ),
    F4 = list(
      name = "F4",
      label = "Schwefel 2.21",
      category = "unimodal",
      fn = .f4,
      lower = -100,
      upper = 100,
      fixed_dim = NA_integer_
    ),
    F5 = list(
      name = "F5",
      label = "Rosenbrock",
      category = "unimodal",
      fn = .f5,
      lower = -30,
      upper = 30,
      fixed_dim = NA_integer_
    ),
    F6 = list(
      name = "F6",
      label = "Step",
      category = "unimodal",
      fn = .f6,
      lower = -100,
      upper = 100,
      fixed_dim = NA_integer_
    ),
    F7 = list(
      name = "F7",
      label = "Quartic Noise",
      category = "unimodal",
      fn = .f7,
      lower = -1.28,
      upper = 1.28,
      fixed_dim = NA_integer_
    ),
    F8 = list(
      name = "F8",
      label = "Schwefel",
      category = "multimodal",
      fn = .f8,
      lower = -500,
      upper = 500,
      fixed_dim = NA_integer_
    ),
    F9 = list(
      name = "F9",
      label = "Rastrigin",
      category = "multimodal",
      fn = .f9,
      lower = -5.12,
      upper = 5.12,
      fixed_dim = NA_integer_
    ),
    F10 = list(
      name = "F10",
      label = "Ackley",
      category = "multimodal",
      fn = .f10,
      lower = -32,
      upper = 32,
      fixed_dim = NA_integer_
    ),
    F11 = list(
      name = "F11",
      label = "Griewank",
      category = "multimodal",
      fn = .f11,
      lower = -600,
      upper = 600,
      fixed_dim = NA_integer_
    ),
    F12 = list(
      name = "F12",
      label = "Penalized 1",
      category = "multimodal",
      fn = .f12,
      lower = -50,
      upper = 50,
      fixed_dim = NA_integer_
    ),
    F13 = list(
      name = "F13",
      label = "Penalized 2",
      category = "multimodal",
      fn = .f13,
      lower = -50,
      upper = 50,
      fixed_dim = NA_integer_
    ),
    F14 = list(
      name = "F14",
      label = "Shekel's Foxholes",
      category = "fixed_dimension",
      fn = .f14,
      lower = -65.536,
      upper = 65.536,
      fixed_dim = 2L
    ),
    F15 = list(
      name = "F15",
      label = "Kowalik",
      category = "fixed_dimension",
      fn = .f15,
      lower = -5,
      upper = 5,
      fixed_dim = 4L
    ),
    F16 = list(
      name = "F16",
      label = "Six-Hump Camel Back",
      category = "fixed_dimension",
      fn = .f16,
      lower = -5,
      upper = 5,
      fixed_dim = 2L
    ),
    F17 = list(
      name = "F17",
      label = "Branin",
      category = "fixed_dimension",
      fn = .f17,
      lower = c(-5, 0),
      upper = c(10, 15),
      fixed_dim = 2L
    ),
    F18 = list(
      name = "F18",
      label = "Goldstein-Price",
      category = "fixed_dimension",
      fn = .f18,
      lower = -2,
      upper = 2,
      fixed_dim = 2L
    ),
    F19 = list(
      name = "F19",
      label = "Hartman 3",
      category = "fixed_dimension",
      fn = .f19,
      lower = 0,
      upper = 1,
      fixed_dim = 3L
    ),
    F20 = list(
      name = "F20",
      label = "Hartman 6",
      category = "fixed_dimension",
      fn = .f20,
      lower = 0,
      upper = 1,
      fixed_dim = 6L
    ),
    F21 = list(
      name = "F21",
      label = "Shekel 5",
      category = "fixed_dimension",
      fn = .f21,
      lower = 0,
      upper = 10,
      fixed_dim = 4L
    ),
    F22 = list(
      name = "F22",
      label = "Shekel 7",
      category = "fixed_dimension",
      fn = .f22,
      lower = 0,
      upper = 10,
      fixed_dim = 4L
    ),
    F23 = list(
      name = "F23",
      label = "Shekel 10",
      category = "fixed_dimension",
      fn = .f23,
      lower = 0,
      upper = 10,
      fixed_dim = 4L
    )
  )
}

#' List built-in benchmark functions
#'
#' Returns a summary table of the built-in F1-F23 benchmark functions.
#'
#' @return A data frame.
#' @examples
#' list_benchmarks()
#' @export
list_benchmarks <- function() {
  reg <- .benchmark_registry()

  data.frame(
    name = names(reg),
    label = vapply(reg, function(z) z$label, character(1)),
    category = vapply(reg, function(z) z$category, character(1)),
    fixed_dim = vapply(
      reg,
      function(z) if (is.na(z$fixed_dim)) "scalable" else as.character(z$fixed_dim),
      character(1)
    ),
    row.names = NULL,
    stringsAsFactors = FALSE
  )
}

#' Get a built-in benchmark definition
#'
#' Returns a built-in benchmark definition by name.
#'
#' @param name Benchmark name such as \code{"F1"} or \code{"F14"}.
#'
#' @return A list containing the benchmark function and its metadata.
#' @examples
#' b <- get_benchmark("F1")
#' b$fn(rep(0, 5))
#' @export
get_benchmark <- function(name) {
  reg <- .benchmark_registry()

  if (!is.character(name) || length(name) != 1 || !name %in% names(reg)) {
    stop("Unknown benchmark name. Use list_benchmarks() to see available benchmarks.")
  }

  reg[[name]]
}

#' Unpack MLP parameter vector
#'
#' Converts a flat parameter vector into weight matrices and bias vectors
#' for a single-hidden-layer multilayer perceptron.
#'
#' @param params Numeric parameter vector.
#' @param input_dim Number of input variables.
#' @param hidden_dim Number of hidden neurons.
#' @param output_dim Number of output neurons.
#'
#' @return A list containing \code{W1}, \code{b1}, \code{W2}, and \code{b2}.
unpack_mlp_params <- function(params, input_dim, hidden_dim, output_dim) {
  idx1 <- input_dim * hidden_dim
  idx2 <- idx1 + hidden_dim
  idx3 <- idx2 + hidden_dim * output_dim

  W1 <- matrix(params[1:idx1], nrow = input_dim, ncol = hidden_dim)
  b1 <- params[(idx1 + 1):idx2]
  W2 <- matrix(params[(idx2 + 1):idx3], nrow = hidden_dim, ncol = output_dim)
  b2 <- params[(idx3 + 1):(idx3 + output_dim)]

  list(
    W1 = W1,
    b1 = b1,
    W2 = W2,
    b2 = b2
  )
}

#' Forward pass for a single-hidden-layer MLP
#'
#' Computes hidden-layer activations and output predictions.
#'
#' @param params Numeric parameter vector.
#' @param X Input matrix.
#' @param input_dim Number of input variables.
#' @param hidden_dim Number of hidden neurons.
#' @param output_dim Number of output neurons.
#'
#' @return A list containing hidden activations and predictions.
forward_mlp <- function(params, X, input_dim, hidden_dim, output_dim) {
  net <- unpack_mlp_params(params, input_dim, hidden_dim, output_dim)

  Z1 <- X %*% net$W1 + matrix(net$b1, nrow = nrow(X), ncol = length(net$b1), byrow = TRUE)
  A1 <- sigmoid(Z1)

  Z2 <- A1 %*% net$W2 + matrix(net$b2, nrow = nrow(A1), ncol = length(net$b2), byrow = TRUE)
  Y_hat <- sigmoid(Z2)

  list(
    A1 = A1,
    Y_hat = Y_hat
  )
}

#' MSE fitness function for MLP training
#'
#' Computes mean squared error for a single-hidden-layer MLP.
#'
#' @param params Numeric parameter vector.
#' @param X Input matrix.
#' @param Y Output matrix.
#' @param input_dim Number of input variables.
#' @param hidden_dim Number of hidden neurons.
#' @param output_dim Number of output neurons.
#'
#' @return Mean squared error.
mlp_mse_fitness <- function(params, X, Y, input_dim, hidden_dim, output_dim) {
  pred <- forward_mlp(params, X, input_dim, hidden_dim, output_dim)$Y_hat
  mean((Y - pred)^2)
}

#' Secretary Bird Optimization Algorithm
#'
#' General-purpose continuous optimization using the Secretary Bird
#' Optimization Algorithm (SBOA).
#'
#' @param fn Objective function to be minimized.
#' @param lower Lower bounds for decision variables.
#' @param upper Upper bounds for decision variables.
#' @param n_agents Number of search agents.
#' @param max_iter Maximum number of iterations.
#' @param ... Additional arguments passed to \code{fn}.
#' @param verbose Logical; if \code{TRUE}, progress is printed.
#' @param seed Optional random seed.
#'
#' @return An object of class \code{"sboa"}.
#' @references
#' Fu, W., Wang, K., Liu, J., et al. (2024). Secretary Bird Optimization Algorithm.
#' Artificial Intelligence Review. DOI: 10.1007/s10462-024-10729-y
#' @importFrom stats runif rnorm
#' @examples
#' sphere <- function(x) sum(x^2)
#'
#' res <- sboa(
#'   fn = sphere,
#'   lower = rep(-10, 5),
#'   upper = rep(10, 5),
#'   n_agents = 10,
#'   max_iter = 20,
#'   seed = 123,
#'   verbose = FALSE
#' )
#'
#' print(res)
#' @export
sboa <- function(fn,
                 lower,
                 upper,
                 n_agents = 30,
                 max_iter = 500,
                 ...,
                 verbose = TRUE,
                 seed = NULL) {

  if (!is.null(seed)) {
    set.seed(seed)
  }

  lower <- as.numeric(lower)
  upper <- as.numeric(upper)

  if (length(lower) != length(upper)) {
    stop("lower and upper must have the same length.")
  }

  dimension <- length(lower)

  X <- matrix(NA_real_, nrow = n_agents, ncol = dimension)
  for (j in seq_len(dimension)) {
    X[, j] <- runif(n_agents, min = lower[j], max = upper[j])
  }

  fit <- vapply(seq_len(n_agents), function(i) fn(X[i, ], ...), numeric(1))

  best_idx <- which.min(fit)
  best_pos <- X[best_idx, ]
  best_score <- fit[best_idx]
  convergence <- numeric(max_iter)

  for (t in seq_len(max_iter)) {
    CF <- (1 - t / max_iter)^(2 * t / max_iter)

    current_best_idx <- which.min(fit)
    current_best <- fit[current_best_idx]

    if (current_best < best_score) {
      best_score <- current_best
      best_pos <- X[current_best_idx, ]
    }

    # Predation strategy
    for (i in seq_len(n_agents)) {
      if (t < max_iter / 3) {
        ids <- sample(seq_len(n_agents), 2, replace = FALSE)
        r1 <- runif(1)
        x_new <- X[i, ] + (X[ids[1], ] - X[ids[2], ]) * r1

      } else if (t < 2 * max_iter / 3) {
        rb <- rnorm(dimension)
        x_new <- best_pos + exp((t / max_iter)^4) * (rb - 0.5) * (best_pos - X[i, ])

      } else {
        rl <- 0.5 * levy_flight(dimension)
        x_new <- best_pos + CF * X[i, ] * rl
      }

      x_new <- clip_bounds(x_new, lower, upper)
      f_new <- fn(x_new, ...)

      if (f_new <= fit[i]) {
        X[i, ] <- x_new
        fit[i] <- f_new
      }
    }

    # Escape strategy
    r <- runif(1)
    k <- sample(seq_len(n_agents), 1)
    x_random <- X[k, ]

    for (i in seq_len(n_agents)) {
      if (r < 0.5) {
        rb <- runif(dimension, -1, 1)
        x_new <- best_pos + (1 - t / max_iter)^2 * (2 * rb - 1) * X[i, ]
      } else {
        K <- round(1 + runif(1))
        r2 <- runif(dimension)
        x_new <- X[i, ] + r2 * (x_random - K * X[i, ])
      }

      x_new <- clip_bounds(x_new, lower, upper)
      f_new <- fn(x_new, ...)

      if (f_new <= fit[i]) {
        X[i, ] <- x_new
        fit[i] <- f_new
      }
    }

    best_idx <- which.min(fit)
    if (fit[best_idx] < best_score) {
      best_score <- fit[best_idx]
      best_pos <- X[best_idx, ]
    }

    convergence[t] <- best_score

    if (verbose) {
      message("Iteration ", t, " | Best Cost = ", signif(best_score, 6))
    }
  }

  out <- list(
    par = best_pos,
    value = best_score,
    convergence = convergence,
    population = X,
    fitness = fit,
    call = match.call()
  )

  class(out) <- "sboa"
  out
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{mape_vec}
\alias{mape_vec}
\title{Mean absolute percentage error}
\usage{
mape_vec(actual, predicted, eps = 1e-08)
}
\arguments{
\item{actual}{Numeric vector or matrix of observed values.}

\item{predicted}{Numeric vector or matrix of predicted values.}

\item{eps}{Small constant to avoid division by zero.}
}
\value{
A numeric scalar.
}
\description{
Computes the mean absolute percentage error between actual and predicted values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sboa_mlp.R
\name{plot.sboa_mlp}
\alias{plot.sboa_mlp}
\title{Plot method for SBOA-MLP objects}
\usage{
\method{plot}{sboa_mlp}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"sboa_mlp"}.}

\item{...}{Additional graphical arguments passed to \code{graphics::plot()}.}
}
\value{
No return value. Called for its side effect.
}
\description{
Plots the convergence curve of a trained SBOA-MLP model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{clip_bounds}
\alias{clip_bounds}
\title{Clip values to lower and upper bounds}
\usage{
clip_bounds(x, lower, upper)
}
\arguments{
\item{x}{Numeric vector.}

\item{lower}{Lower bound vector.}

\item{upper}{Upper bound vector.}
}
\value{
A bounded numeric vector.
}
\description{
Restricts each element of a vector to the given bounds.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{levy_flight}
\alias{levy_flight}
\title{Lévy flight step generator}
\usage{
levy_flight(d, beta = 1.5)
}
\arguments{
\item{d}{Dimension of the step vector.}

\item{beta}{Stability parameter.}
}
\value{
A numeric vector of length \code{d}.
}
\description{
Generates Lévy-distributed random steps.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sboa_mlp.R
\name{print.sboa_mlp}
\alias{print.sboa_mlp}
\title{Print method for SBOA-MLP objects}
\usage{
\method{print}{sboa_mlp}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"sboa_mlp"}.}

\item{...}{Additional arguments, ignored.}
}
\value{
The input object, invisibly.
}
\description{
Prints a summary of a trained SBOA-MLP model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{sigmoid}
\alias{sigmoid}
\title{Sigmoid activation function}
\usage{
sigmoid(x)
}
\arguments{
\item{x}{A numeric vector, matrix, or scalar.}
}
\value{
Transformed values in the interval (0, 1).
}
\description{
Computes the sigmoid transformation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp_helpers.R
\name{mlp_mse_fitness}
\alias{mlp_mse_fitness}
\title{MSE fitness function for MLP training}
\usage{
mlp_mse_fitness(params, X, Y, input_dim, hidden_dim, output_dim)
}
\arguments{
\item{params}{Numeric parameter vector.}

\item{X}{Input matrix.}

\item{Y}{Output matrix.}

\item{input_dim}{Number of input variables.}

\item{hidden_dim}{Number of hidden neurons.}

\item{output_dim}{Number of output neurons.}
}
\value{
Mean squared error.
}
\description{
Computes mean squared error for a single-hidden-layer MLP.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{apply_minmax}
\alias{apply_minmax}
\title{Apply min-max normalization using existing bounds}
\usage{
apply_minmax(x, mins, maxs)
}
\arguments{
\item{x}{Numeric vector or matrix.}

\item{mins}{Column minima.}

\item{maxs}{Column maxima.}
}
\value{
Scaled matrix.
}
\description{
Apply min-max normalization using existing bounds
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sboa_mlp.R
\name{sboa_mlp}
\alias{sboa_mlp}
\title{Train a single-hidden-layer MLP using SBOA}
\usage{
sboa_mlp(
  X_train,
  y_train,
  hidden_dim = 10,
  n_agents = 30,
  max_iter = 500,
  lower = -1,
  upper = 1,
  seed = NULL,
  verbose = TRUE
)
}
\arguments{
\item{X_train}{Training input data.}

\item{y_train}{Training output data.}

\item{hidden_dim}{Number of hidden neurons.}

\item{n_agents}{Number of search agents.}

\item{max_iter}{Maximum number of iterations.}

\item{lower}{Lower bound for parameter search.}

\item{upper}{Upper bound for parameter search.}

\item{seed}{Optional random seed.}

\item{verbose}{Logical; if \code{TRUE}, progress is printed.}
}
\value{
An object of class \code{"sboa_mlp"}.
}
\description{
Trains a single-hidden-layer multilayer perceptron with the
Secretary Bird Optimization Algorithm.
}
\examples{
set.seed(123)

X_train <- matrix(runif(40), nrow = 10, ncol = 4)
y_train <- matrix(runif(10), nrow = 10, ncol = 1)

fit <- sboa_mlp(
  X_train = X_train,
  y_train = y_train,
  hidden_dim = 3,
  n_agents = 10,
  max_iter = 20,
  lower = -1,
  upper = 1,
  seed = 123,
  verbose = FALSE
)

print(fit)
pred <- predict(fit, X_train)
head(pred)
}
\references{
Dilber, B., and Ozdemir, A. F. (2026). A novel approach to training
feed-forward multi-layer perceptrons with recently proposed secretary bird
optimization algorithm. Neural Computing and Applications.
DOI: 10.1007/s00521-026-11874-x
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sboa_mlp.R
\name{predict.sboa_mlp}
\alias{predict.sboa_mlp}
\title{Predict method for SBOA-MLP objects}
\usage{
\method{predict}{sboa_mlp}(object, newdata, ...)
}
\arguments{
\item{object}{An object of class \code{"sboa_mlp"}.}

\item{newdata}{New predictor data.}

\item{...}{Additional arguments, ignored.}
}
\value{
A matrix of predicted values.
}
\description{
Generates predictions from a trained SBOA-MLP model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmarks.R
\name{get_benchmark}
\alias{get_benchmark}
\title{Get a built-in benchmark definition}
\usage{
get_benchmark(name)
}
\arguments{
\item{name}{Benchmark name such as \code{"F1"} or \code{"F14"}.}
}
\value{
A list containing the benchmark function and its metadata.
}
\description{
Returns a built-in benchmark definition by name.
}
\examples{
b <- get_benchmark("F1")
b$fn(rep(0, 5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp_helpers.R
\name{unpack_mlp_params}
\alias{unpack_mlp_params}
\title{Unpack MLP parameter vector}
\usage{
unpack_mlp_params(params, input_dim, hidden_dim, output_dim)
}
\arguments{
\item{params}{Numeric parameter vector.}

\item{input_dim}{Number of input variables.}

\item{hidden_dim}{Number of hidden neurons.}

\item{output_dim}{Number of output neurons.}
}
\value{
A list containing \code{W1}, \code{b1}, \code{W2}, and \code{b2}.
}
\description{
Converts a flat parameter vector into weight matrices and bias vectors
for a single-hidden-layer multilayer perceptron.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp_helpers.R
\name{forward_mlp}
\alias{forward_mlp}
\title{Forward pass for a single-hidden-layer MLP}
\usage{
forward_mlp(params, X, input_dim, hidden_dim, output_dim)
}
\arguments{
\item{params}{Numeric parameter vector.}

\item{X}{Input matrix.}

\item{input_dim}{Number of input variables.}

\item{hidden_dim}{Number of hidden neurons.}

\item{output_dim}{Number of output neurons.}
}
\value{
A list containing hidden activations and predictions.
}
\description{
Computes hidden-layer activations and output predictions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{normalize_minmax}
\alias{normalize_minmax}
\title{Min-max normalization}
\usage{
normalize_minmax(x)
}
\arguments{
\item{x}{Numeric vector or matrix.}
}
\value{
A list containing scaled data and min-max values.
}
\description{
Scales each column of a matrix to the interval from 0 to 1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sboa.R
\name{sboa}
\alias{sboa}
\title{Secretary Bird Optimization Algorithm}
\usage{
sboa(
  fn,
  lower,
  upper,
  n_agents = 30,
  max_iter = 500,
  ...,
  verbose = TRUE,
  seed = NULL
)
}
\arguments{
\item{fn}{Objective function to be minimized.}

\item{lower}{Lower bounds for decision variables.}

\item{upper}{Upper bounds for decision variables.}

\item{n_agents}{Number of search agents.}

\item{max_iter}{Maximum number of iterations.}

\item{...}{Additional arguments passed to \code{fn}.}

\item{verbose}{Logical; if \code{TRUE}, progress is printed.}

\item{seed}{Optional random seed.}
}
\value{
An object of class \code{"sboa"}.
}
\description{
General-purpose continuous optimization using the Secretary Bird
Optimization Algorithm (SBOA).
}
\examples{
sphere <- function(x) sum(x^2)

res <- sboa(
  fn = sphere,
  lower = rep(-10, 5),
  upper = rep(10, 5),
  n_agents = 10,
  max_iter = 20,
  seed = 123,
  verbose = FALSE
)

print(res)
}
\references{
Fu, W., Wang, K., Liu, J., et al. (2024). Secretary Bird Optimization Algorithm.
Artificial Intelligence Review. DOI: 10.1007/s10462-024-10729-y
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{mae_vec}
\alias{mae_vec}
\title{Mean absolute error}
\usage{
mae_vec(actual, predicted)
}
\arguments{
\item{actual}{Numeric vector or matrix of observed values.}

\item{predicted}{Numeric vector or matrix of predicted values.}
}
\value{
A numeric scalar.
}
\description{
Computes the mean absolute error between actual and predicted values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sboa_methods.R
\name{plot.sboa}
\alias{plot.sboa}
\title{Plot method for SBOA objects}
\usage{
\method{plot}{sboa}(x, ...)
}
\arguments{
\item{x}{An object of class \code{"sboa"}.}

\item{...}{Additional graphical arguments passed to \code{plot()}.}
}
\value{
No return value. Called for its side effect.
}
\description{
Plots the convergence curve of an SBOA optimization result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmarks.R
\name{list_benchmarks}
\alias{list_benchmarks}
\title{List built-in benchmark functions}
\usage{
list_benchmarks()
}
\value{
A data frame.
}
\description{
Returns a summary table of the built-in F1-F23 benchmark functions.
}
\examples{
list_benchmarks()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{denormalize_minmax}
\alias{denormalize_minmax}
\title{Reverse min-max normalization}
\usage{
denormalize_minmax(x_scaled, mins, maxs)
}
\arguments{
\item{x_scaled}{Scaled numeric vector or matrix.}

\item{mins}{Column minima.}

\item{maxs}{Column maxima.}
}
\value{
Values on the original scale.
}
\description{
Reverse min-max normalization
}
